module IEval where

import Data.Map(Map)
import qualified Data.Map as M
import Data.Char(ord, chr, isPrint)
import Control.Monad.State.Strict
import System.IO
import System.IO.Error
import Text.Printf

import System.Environment   
import Control.Exception

import Absy
import Parser

-- State of the BF machine
data S = S { mem :: Map Int Char 
           , ptr :: Int 
           }

initial :: S
initial = S { mem = M.empty, ptr = 0}

getPtr :: S -> Char
getPtr s = M.findWithDefault '\0' (ptr s) (mem s)

co :: (Int-> Int)-> Char-> Char
co m c = chr (m (ord c) `mod` 255)

-- The BF monad transforms the state of the BF machine, and can do some IO.
type BF a = StateT S IO a

evalc :: Cmd-> BF ()
evalc FWD = modify $ \s-> s{ptr = ptr s+ 1}
evalc BWD = modify $ \s-> s{ptr = ptr s- 1}
evalc INC = modify $ \s-> s{mem = M.insert (ptr s) (co (+ 1) (getPtr s)) (mem s)}
evalc DEC = modify $ \s-> s{mem = M.insert (ptr s) (co (subtract 1) (getPtr s)) (mem s)}
evalc INP = do
    c <- liftIO $ getChar
    modify $ \s-> s{mem = M.insert (ptr s) c (mem s)}
evalc OUT = do 
    c<- gets getPtr 
    liftIO $ putChar c 
evalc (WHL cs) = do 
    p<- gets getPtr
    when (p /= '\0') $ do eval cs; evalc (WHL cs)
      
eval :: [Cmd]-> BF ()
eval = mapM_ evalc

run :: FilePath-> IO ()
run filenm = do
  pr <- parseFile filenm
  case pr of 
    Right cs -> (do
      hSetBuffering stdin NoBuffering -- This does nothing under windows
      s <- execStateT (eval cs) initial
      putStrLn $ "\n\nEXECUTION TERMINATED.\nMEMORY DUMP:\n" ++ memdump (mem s)
      ) `catch` (\e-> putStrLn $ "EXECUTION ABORTED. UNEXPECTED I/O ERROR: "++ show (e :: IOException))
    Left e -> putStrLn $ "*** ERROR: "++ e 
  

memdump :: Map Int Char-> String
memdump mem = 
  let ks = M.keys mem
      s = (head ks `div` 8) * 8
      dump x | x > last ks = ""
             | otherwise = oneline x ++ dump (x+8)
      oneline x = printf "%04x %02x %02x %02x %02x %02x %02x %02x %02x " 
                         x (l x) (l $ x+1) (l $ x+2) (l $ x+3)  (l $ x+4) (l $ x+5) (l $ x+6) (l $ x+7)
                  ++ map s [x .. x+7]++ "\n"                   
                    where l x = M.findWithDefault (chr 0) x mem
                          s x = if isPrint (l x) then l x else '.'
  in  if null ks then "---- NO MEMORY WRITTEN" else dump s

main :: IO ()
main = do
  putStrLn "Welcome to Brainf***"
  p <- getProgName
  args <- getArgs
  case args of 
    [f] -> do run f
    _   -> do putStrLn $ "*** ERROR: Usage "++ show p++ " <filename>"

